#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
# This file is a part of debmutate.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for debmutate.deb822."""

from . import (
    TestCase,
    TestCaseInTempDir,
    )

from debian.deb822 import Deb822

from debmutate.deb822 import (
    ChangeConflict,
    Deb822Editor,
    dump_paragraphs,
    reformat_deb822,
    )
from debmutate.reformatting import (
    GeneratedFile,
    )


class ReformatDeb822Tests(TestCase):

    def test_comment(self):
        text_with_comment = b"""\
Source: blah
# A comment
Testsuite: autopkgtest

"""
        self.assertEqual(
            reformat_deb822(text_with_comment), text_with_comment)

    def test_fine(self):
        self.assertTrue(reformat_deb822(b"""\
Source: blah
Testsuite: autopkgtest

"""), b"""\
Source: blah
Testsuite: autogpktest

""")


class DumpParagraphsTests(TestCase):

    def test_simple(self):
        self.assertEqual(dump_paragraphs([Deb822({
                'Source': 'blah',
                'Testsuite': 'autopkgtest'
            })]), b"""\
Source: blah
Testsuite: autopkgtest
""")

    def test_multi(self):
        self.assertEqual(dump_paragraphs([
            Deb822({
                'Source': 'blah',
                'Testsuite': 'autopkgtest'
            }),
            Deb822({'Package': 'bloe'})]), b"""\
Source: blah
Testsuite: autopkgtest

Package: bloe
""")


class UpdateControlTests(TestCaseInTempDir):

    def test_do_not_edit(self):
        self.build_tree_contents([('controlfile', """\
# DO NOT EDIT
# This file was generated by blah

Source: blah
Testsuite: autopkgtest

""")])

        def change():
            with Deb822Editor('controlfile') as updater:
                for c in updater.paragraphs:
                    c['Source'] = 'blah1'
        self.assertRaises(GeneratedFile, change)

    def test_do_not_edit_force(self):
        self.build_tree_contents([('controlfile', """\
Source: blah
Testsuite: autopkgtest

"""), ('controlfile.in', 'bar')])

        with Deb822Editor('controlfile', allow_generated=True) as updater:
            for c in updater.paragraphs:
                c['Source'] = 'blah1'

    def test_do_not_edit_no_change(self):
        self.build_tree_contents([('controlfile', """\
# DO NOT EDIT
# This file was generated by blah

Source: blah
Testsuite: autopkgtest

""")])
        with Deb822Editor('controlfile'):
            pass

    def test_unpreservable(self):
        self.build_tree_contents([('controlfile', """\
Source: blah
# A comment
Testsuite: autopkgtest

""")])

        with Deb822Editor('controlfile') as updater:
            for control in updater.paragraphs:
                control["NewField"] = "New Field"

    def test_modify_paragraph(self):
        self.build_tree_contents([('controlfile', """\
Source: blah
Testsuite: autopkgtest
""")])

        with Deb822Editor('controlfile') as updater:
            for control in updater.paragraphs:
                control["XS-Vcs-Git"] = "git://github.com/example/example"
        self.assertTrue(updater.changed)
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest
XS-Vcs-Git: git://github.com/example/example
""", 'controlfile')

    def test_doesnt_strip_whitespace(self):
        self.build_tree_contents([('controlfile', """\
Source: blah
Testsuite: autopkgtest

""")])
        with Deb822Editor('controlfile') as updater:
            pass
        self.assertFalse(updater.changed)
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest

""", 'controlfile')


class ApplyChangesTests(TestCaseInTempDir):

    def setUp(self):
        super().setUp()
        self.build_tree_contents([('controlfile', """\
Source: blah
Testsuite: autopkgtest

""")])

    def test_simple_set(self):
        with Deb822Editor('controlfile') as updater:
            updater.apply_changes(
                {('Source', 'blah'): [('Build-Depends', None, 'foo')]})
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest
Build-Depends: foo
""", 'controlfile', strip_trailing_whitespace=True)

    def test_simple_change(self):
        with Deb822Editor('controlfile') as updater:
            updater.apply_changes(
                {('Source', 'blah'): [('Testsuite', 'autopkgtest', 'foo')]})
        self.assertFileEqual("""\
Source: blah
Testsuite: foo
""", 'controlfile', strip_trailing_whitespace=True)

    def test_change_conflict(self):
        with Deb822Editor('controlfile') as updater:
            self.assertRaises(
                ChangeConflict, updater.apply_changes,
                {('Source', 'blah'): [('Testsuite', 'different', 'foo')]})

    def test_simple_delete(self):
        with Deb822Editor('controlfile') as updater:
            updater.apply_changes(
                {('Source', 'blah'): [('Testsuite', 'autopkgtest', None)]})
        self.assertFileEqual("""\
Source: blah
""", 'controlfile', strip_trailing_whitespace=True)

    def test_delete_conflict(self):
        with Deb822Editor('controlfile') as updater:
            self.assertRaises(
                ChangeConflict, updater.apply_changes,
                {('Source', 'blah'): [('Nonexistent', 'autopkgtest', None)]})

    def test_simple_add_para(self):
        with Deb822Editor('controlfile') as updater:
            updater.apply_changes(
                {('Source', 'new'): [
                    ('Source', None, 'new'),
                    ('Build-Depends', None, 'bar')]})
        self.assertFileEqual("""\
Source: blah
Testsuite: autopkgtest

Source: new
Build-Depends: bar
""", 'controlfile')

    def test_simple_add_para_conflict(self):
        with Deb822Editor('controlfile') as updater:
            self.assertRaises(
                ChangeConflict,
                updater.apply_changes,
                {('Source', 'new'): [
                    ('Source', None, 'new'),
                    ('Build-Depends', 'bar', 'bar')]})

    def test_sort_paragaphs(self):
        with Deb822Editor('controlfile') as updater:
            try:
                updater.sort_paragraphs(sort_key=lambda m: m['Source'])
            except NotImplementedError:
                # Version of python-debian too old
                pass



#!/usr/bin/python3
# Copyright (C) 2018-2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the vcs module."""

from unittest import TestCase

from debmutate.vcs import (
    VcsUrl,
    split_vcs_url,
    unsplit_vcs_url,
    mangle_version_for_git,
    gbp_expand_tag_name,
    )


class SplitVcsUrlTests(TestCase):

    def test_none(self):
        self.assertEqual(
            VcsUrl('https://github.com/jelmer/example', None, None),
            split_vcs_url('https://github.com/jelmer/example'))
        self.assertEqual(
            VcsUrl(
                'https://github.com/jelmer/example', None,
                'path/to/packaging'),
            split_vcs_url(
                'https://github.com/jelmer/example [path/to/packaging]'))

    def test_branch(self):
        self.assertEqual(
            VcsUrl(
                'https://github.com/jelmer/example', 'master',
                'path/to/packaging'),
            split_vcs_url(
                'https://github.com/jelmer/example [path/to/packaging] '
                '-b master'))
        self.assertEqual(
            VcsUrl(
                'https://github.com/jelmer/example', 'master',
                'path/to/packaging'),
            split_vcs_url(
                'https://github.com/jelmer/example -b master '
                '[path/to/packaging]'))
        self.assertEqual(
            VcsUrl('https://github.com/jelmer/example', 'master', None),
            split_vcs_url(
                'https://github.com/jelmer/example -b master'))


class UnsplitVcsUrlTests(TestCase):

    def test_none(self):
        self.assertEqual(
            'https://github.com/jelmer/example',
            unsplit_vcs_url('https://github.com/jelmer/example', None, None))
        self.assertEqual(
            'https://github.com/jelmer/example [path/to/packaging]',
            unsplit_vcs_url(
                'https://github.com/jelmer/example', None,
                'path/to/packaging'))

    def test_branch(self):
        self.assertEqual(
            'https://github.com/jelmer/example -b master '
            '[path/to/packaging]',
            unsplit_vcs_url(
                'https://github.com/jelmer/example', 'master',
                'path/to/packaging'))
        self.assertEqual(
            'https://github.com/jelmer/example -b master',
            unsplit_vcs_url(
                'https://github.com/jelmer/example', 'master', None))


class MangleVersionForGitTests(TestCase):

    def test_replace_tilde(self):
        self.assertEqual('1.0_1', mangle_version_for_git('1.0~1'))

    def test_normal(self):
        self.assertEqual('1.0', mangle_version_for_git('1.0'))


class ExpandGbpTagFormatTests(TestCase):

    def test_gbp_tag_format(self):
        self.assertEqual(
            'blah-1.0', gbp_expand_tag_name('blah-%(version)s', '1.0'))
        self.assertEqual(
            'blah-0.1-1',
            gbp_expand_tag_name('blah-%(version%~%-)s', '0.1~1'))



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij
# This file is a part of debmutate.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with control files."""

__all__ = [
    'dh_gnome_clean',
    'pg_buildext_updatecontrol',
    'guess_template_type',
    'ControlEditor',
    'update_control',
    'parse_relations',
    'format_relations',
    'get_relation',
    'iter_relations',
    'ensure_minimum_version',
    'ensure_exact_version',
    'ensure_some_version',
    'filter_dependencies',
    'drop_dependency',
    'delete_from_list',
    'is_dep_implied',
    'is_relation_implied',
    'parse_standards_version',
    'PkgRelationFieldEditor',
    ]

import collections
import contextlib
from itertools import takewhile
import operator
import os
import re
import time
from typing import (
    ContextManager,
    Optional,
    Callable,
    Tuple,
    Union,
    List,
    Iterable,
    Dict,
)

import subprocess
import warnings

from debian.changelog import Version

from ._deb822 import PkgRelation
from .deb822 import (
    Deb822Editor,
    ChangeConflict,
    Deb822Paragraph,
    parse_deb822_file,
    )
from .reformatting import GeneratedFile


# TODO(jelmer): dedupe with scripts/wrap-and-sort in devscripts
CONTROL_LIST_FIELDS = (
    "Breaks",
    "Build-Conflicts",
    "Build-Conflicts-Arch",
    "Build-Conflicts-Indep",
    "Build-Depends",
    "Build-Depends-Arch",
    "Build-Depends-Indep",
    "Built-Using",
    "Conflicts",
    "Depends",
    "Enhances",
    "Pre-Depends",
    "Provides",
    "Recommends",
    "Replaces",
    "Suggests",
    "Xb-Npp-MimeType",
)


class MissingSourceParagraph(Exception):
    """The first paragraph is not a Source paragraph."""


class TemplateExpansionFailed(Exception):

    def __init__(self, command, stderr):
        self.command = command
        self.stderr = stderr
        super().__init__(
            "Template expansion ({!r}) failed: {}".format(command, stderr))


class TemplateExpandCommandMissing(Exception):

    def __init__(self, command):
        self.command = command
        super().__init__(
            "Command for expanding template file missing: %s" %
            command)


def parse_relation(t: str):
    with warnings.catch_warnings():
        suppress_substvar_warnings()
        return PkgRelation.parse(t)

# From scripts/wrap-and-sort in devscripts
# Copyright (C) 2010-2018, Benjamin Drung <bdrung@debian.org>
#               2010, Stefano Rivera <stefanor@ubuntu.com>


def _sort_packages_key(package):
    # Sort dependencies starting with a "real" package name before ones
    # starting with a substvar
    return 0 if re.match("[a-z0-9]", package) else 1, package


def _wrap_field(control, entry, sort, formatter):
    from debian._deb822_repro import LIST_COMMA_SEPARATED_INTERPRETATION
    view = control.as_interpreted_dict_view(
        LIST_COMMA_SEPARATED_INTERPRETATION)
    with view[entry] as field_content:
        seen = set()
        for package_ref in field_content.iter_value_references():
            value = package_ref.value
            new_value = " | ".join(x.strip() for x in value.split("|"))
            if not sort or new_value not in seen:
                package_ref.value = new_value
                seen.add(new_value)
            else:
                package_ref.remove()
        if sort:
            field_content.sort(key=_sort_packages_key)
            if formatter:
                field_content.value_formatter(formatter)
            field_content.reformat_when_finished()


# End import from devscripts


def dh_gnome_clean(path: str = '.') -> None:
    """Run the dh_gnome_clean command.

    This needs to do some post-hoc cleaning, since dh_gnome_clean
    writes various debhelper log files that should not be checked in.

    Args:
      path: Path to run dh_gnome_clean in
    """
    for n in os.listdir(os.path.join(path, 'debian')):
        if n.endswith('.debhelper.log'):
            raise AssertionError('pre-existing .debhelper.log files')
    if not os.path.exists(os.path.join(path, 'debian/changelog')):
        raise AssertionError('no changelog file in %s' % path)
    try:
        subprocess.check_call(["dh_gnome_clean"], cwd=path)
    except FileNotFoundError as e:
        raise TemplateExpandCommandMissing("dh_gnome_clean") from e
    for n in os.listdir(os.path.join(path, 'debian')):
        if n.endswith('.debhelper.log'):
            os.unlink(os.path.join(path, 'debian', n))


def pg_buildext_updatecontrol(path: str = '.') -> None:
    """Run the 'pg_buildext updatecontrol' command.

    Args:
      path: path to run pg_buildext updatecontrol in
    """
    try:
        subprocess.check_call(["pg_buildext", "updatecontrol"], cwd=path)
    except FileNotFoundError as e:
        raise TemplateExpandCommandMissing("pg_buildext") from e


def guess_template_type(
        template_path: str,
        debian_path: Optional[str] = None) -> Optional[str]:
    """Guess the type for a control template.

    Args:
      template_path: Template path
      debian_path: Path to debian directory
    Returns:
      Name of template type; None if unknown
    """
    # TODO(jelmer): This should use a proper make file parser of some sort..
    if debian_path is not None:
        try:
            with open(os.path.join(debian_path, 'rules'), 'rb') as f:
                for line in f:
                    if line.startswith(b'debian/control:'):
                        return 'rules'
                    if line.startswith(b'debian/%: debian/%.in'):
                        return 'rules'
                    if line.startswith(b'include /usr/share/blends-dev/rules'):
                        return 'rules'
        except FileNotFoundError:
            pass
    try:
        with open(template_path, 'rb') as f:
            template = f.read()
            if b'@GNOME_TEAM@' in template:
                return 'gnome'
            elif b'@cdbs@' in template:
                return 'cdbs'
            elif b'PGVERSION' in template:
                return 'postgresql'
            else:
                try:
                    deb822 = next(iter(
                        parse_deb822_file(
                            template.splitlines(),
                            accept_files_with_error_tokens=True)))
                except StopIteration:
                    build_depends = ''
                else:
                    build_depends = deb822.get('Build-Depends', '')
                if any(iter_relations(build_depends, 'gnome-pkg-tools')):
                    return 'gnome'
                if any(iter_relations(build_depends, 'cdbs')):
                    return 'cdbs'
    except IsADirectoryError:
        return 'directory'
    if (debian_path is not None
            and os.path.exists(os.path.join(debian_path, 'debcargo.toml'))):
        return 'debcargo'
    return None


def _cdbs_resolve_conflict(
        para_key: Tuple[str, str],
        field: str,
        actual_old_value: Optional[str],
        template_old_value: Optional[str],
        actual_new_value: Optional[str]
        ) -> Optional[str]:
    if (para_key[0] == 'Source' and field == 'Build-Depends' and
            template_old_value is not None and
            actual_old_value is not None and
            actual_new_value is not None):
        if actual_old_value in actual_new_value:
            # We're simply adding to the existing list
            return actual_new_value.replace(
                actual_old_value, template_old_value)
        else:
            existing = [v[1] for v in parse_relations(actual_old_value)]
            ret = template_old_value
            for _, v, _ in parse_relations(actual_new_value):
                if any(is_relation_implied(v, r) for r in existing):
                    continue
                ret = ensure_relation(ret, v)
            return ret
    raise ChangeConflict(
        para_key, field, actual_old_value, template_old_value,
        actual_new_value)


def _expand_control_template(
        template_path: str, path: str, template_type: str):
    package_root = os.path.dirname(os.path.dirname(path)) or '.'
    if template_type == 'rules':
        try:
            path_mtime = os.stat(path).st_mtime
        except FileNotFoundError:
            path_mtime = None
        while os.stat(template_path).st_mtime == path_mtime:
            # Wait until mtime has changed, so that make knows to regenerate.
            os.utime(template_path, (time.time(), time.time()))
        try:
            subprocess.check_call(
                ['./debian/rules', 'debian/control'],
                stderr=subprocess.PIPE,
                cwd=package_root)
        except subprocess.CalledProcessError as e:
            raise TemplateExpansionFailed(
                ["./debian/rules"], e.stderr.decode())
    elif template_type == 'gnome':
        dh_gnome_clean(package_root)
    elif template_type == 'postgresql':
        pg_buildext_updatecontrol(package_root)
    elif template_type == 'directory':
        raise GeneratedFile(path, template_path)
    else:
        raise AssertionError


def _update_control_template(
        template_path: str, path: str, changes, expand_template=True):
    template_type = guess_template_type(
            template_path, debian_path=os.path.dirname(path))
    if template_type is None:
        raise GeneratedFile(path, template_path)
    if template_type == 'directory':
        # We can't handle these yet
        raise GeneratedFile(path, template_path)
    with Deb822Editor(
            template_path, accept_files_with_error_tokens=True) as updater:
        resolve_conflict: Optional[Callable[[
            Tuple[str, str], str, Optional[str], Optional[str], Optional[str]],
            Optional[str]]]
        if template_type == 'cdbs':
            resolve_conflict = _cdbs_resolve_conflict
        else:
            resolve_conflict = None
        updater.apply_changes(changes, resolve_conflict=resolve_conflict)
    if not updater.changed:
        # A bit odd, since there were changes to the output file. Anyway.
        return False

    if expand_template:
        if template_type == 'cdbs':
            with Deb822Editor(path, allow_generated=True) as updater:
                updater.apply_changes(changes)
        else:
            _expand_control_template(template_path, path, template_type)
    return True


@contextlib.contextmanager
def _preserve_field_order_preferences(paragraphs):
    description_is_not_last = set()
    for para in paragraphs:
        if 'Package' not in para:
            continue
        if list(para) and list(para)[-1] != 'Description':
            description_is_not_last.add(para['Package'])
    yield
    for para in paragraphs:
        if 'Package' not in para:
            continue
        if para['Package'] not in description_is_not_last:
            # Make sure Description stays the last field
            try:
                para.order_last('Description')
            except KeyError:
                pass  # it's okay if the field doesn't exist


def update_control(path='debian/control', source_package_cb=None,
                   binary_package_cb=None):
    def paragraph_cb(paragraph):
        if paragraph.get("Source"):
            if source_package_cb is not None:
                source_package_cb(paragraph)
        else:
            if binary_package_cb is not None:
                binary_package_cb(paragraph)

    with ControlEditor(path) as updater:
        for paragraph in updater.paragraphs:
            paragraph_cb(paragraph)
    return updater.changed


def _find_template_path(path):
    for template_path in [path + '.in', path + '.m4']:
        if os.path.exists(template_path):
            return template_path
    else:
        return None


ChangesDict = Dict[
    Tuple[str, str], List[Tuple[str, Optional[str], Optional[str]]]]


class ControlEditor:
    """Edit a control file.
    """

    changed: bool
    changed_files: List[str]
    _field_order_preserver: ContextManager

    def __init__(self, path: str = 'debian/control',
                 allow_reformatting: Optional[bool] = None,
                 allow_missing: bool = False):
        self.path = path
        self._primary = Deb822Editor(
            path, allow_reformatting=allow_reformatting,
            allow_missing=allow_missing)
        self._template_only = False

    @classmethod
    def create(cls, path: str = 'debian/control'):
        return cls(path, allow_reformatting=True, allow_missing=True)

    @classmethod
    def from_tree(cls, tree, subpath=None):
        relpath = 'debian/control'
        if subpath not in (None, '.', ''):
            relpath = os.path.join(subpath, relpath)
        return cls(tree.abspath(relpath))

    @property
    def paragraphs(self) -> List[Deb822Paragraph]:
        """List of all the paragraphs."""
        return self._primary.paragraphs

    @property
    def source(self) -> Deb822Paragraph:
        """Source package."""
        for entry in self.paragraphs:
            if not entry.get('Source'):
                raise MissingSourceParagraph()
            return entry
        else:
            p = Deb822Paragraph.new_empty_paragraph()
            self.paragraphs.insert(0, p)
            return p

    @property
    def binaries(self) -> Iterable[Deb822Paragraph]:
        """List of binary packages."""
        for entry in self.paragraphs:
            if entry.get('Package'):
                yield entry

    def changes(self):
        """Return a dictionary describing the changes since the base.

        Returns:
          dictionary mapping tuples of (kind, name) to
            list of (field_name, old_value, new_value)
        """
        orig = self._primary._parse(self._primary._orig_content)
        changes: ChangesDict = {}

        def by_key(ps):
            ret = {}
            for p in ps:
                if not p:
                    continue
                if 'Source' in p:
                    ret[('Source', p['Source'])] = p
                else:
                    ret[('Package', p['Package'])] = p
            return ret

        orig_by_key = by_key(orig)
        new_by_key = by_key(self.paragraphs)
        for key in set(orig_by_key).union(set(new_by_key)):
            old = orig_by_key.get(key, {})
            new = new_by_key.get(key, {})
            if old == new:
                continue
            fields = list(old)
            fields.extend([field for field in new if field not in fields])
            for field in fields:
                if old.get(field) != new.get(field):
                    changes.setdefault(key, []).append(
                        (field, old.get(field), new.get(field)))
        return changes

    def __enter__(self):
        try:
            self._primary.__enter__()
        except FileNotFoundError:
            template_path = _find_template_path(self.path)
            if template_path:
                self._template_only = True
                template_type = guess_template_type(
                    template_path, debian_path=os.path.dirname(self.path))
                if template_type is None:
                    raise GeneratedFile(self.path, template_path)
                _expand_control_template(
                    template_path, self.path, template_type)
                self._primary.__enter__()
            else:
                raise
        self._field_order_preserver = _preserve_field_order_preferences(
            self._primary.paragraphs)
        self._field_order_preserver.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._field_order_preserver.__exit__(exc_type, exc_val, exc_tb)
        try:
            if self._template_only:
                os.unlink(self.path)
                self.changed_files = [self.path]
                raise FileNotFoundError
            self._primary.__exit__(exc_type, exc_val, exc_tb)
        except GeneratedFile as e:
            if not e.template_path:
                raise
            self.changed = _update_control_template(
                e.template_path, self.path, self.changes())
            if self.changed:
                self.changed_files = [e.template_path, self.path]
            else:
                self.changed_files = []
        except FileNotFoundError:
            template_path = _find_template_path(self.path)
            if template_path:
                self.changed = _update_control_template(
                    template_path, self.path, self.changes(),
                    expand_template=not self._template_only)
                if self.changed:
                    self.changed_files = [template_path, self.path]
                else:
                    self.changed_files = []
            else:
                raise
        else:
            self.changed = self._primary.changed
            if self.changed:
                self.changed_files = [self.path]
            else:
                self.changed_files = []
        return False

    def sort_binary_packages(self, keep_first: bool = False) -> None:
        sort_key = operator.itemgetter("Package")
        self._primary.sort_paragraphs(
            skip=1 + int(keep_first), sort_key=sort_key)

    def wrap_and_sort(self, short_indent: bool = False,
                      trailing_comma: bool = False,
                      wrap_always: bool = False,
                      max_line_length: int = 79) -> None:
        """Wrap and sort this control file."""
        try:
            from devscripts.control import wrap_and_sort_formatter
        except ImportError:
            raise NotImplementedError('version of devscripts too old')
        formatter = wrap_and_sort_formatter(
            1 if short_indent else "FIELD_NAME_LENGTH",
            trailing_separator=trailing_comma,
            immediate_empty_line=short_indent,
            max_line_length_one_liner=0 if wrap_always else max_line_length
        )
        for paragraph in self.paragraphs:
            for field in CONTROL_LIST_FIELDS:
                if field in paragraph:
                    _wrap_field(paragraph, field, True, formatter)
            if "Uploaders" in paragraph:
                _wrap_field(paragraph, "Uploaders", False, formatter)
            if "Architecture" in paragraph:
                archs = list(paragraph["Architecture"].split())
                # Sort, with wildcard entries (such as linux-any) first:
                archs = sorted(archs, key=lambda x: ("any" not in x, x))
                paragraph["Architecture"] = " ".join(archs)

    def add_binary(self, contents):
        if isinstance(contents, dict):
            para = Deb822Paragraph.from_dict(contents)
        else:
            para = contents
        return self._primary.paragraphs.append(para)

    def remove(self, para):
        self._primary.paragraphs.remove(para)


def parse_relations(text: str) -> List[Tuple[str, List[PkgRelation], str]]:
    """Parse a package relations string.

    (e.g. a Depends, Provides, Build-Depends, etc field)

    This attemps to preserve some indentation.

    Args:
      text: Text to parse
    Returns: list of tuples with (whitespace, relation, whitespace)
    """
    ret: List[Tuple[str, List[PkgRelation], str]] = []
    for top_level in text.split(','):
        if top_level == "":
            if ',' not in text:
                return []
        if top_level.isspace():
            ret.append((top_level, [], ''))
            continue
        head_whitespace = ''
        for i in range(len(top_level)):
            if not top_level[i].isspace():
                if i > 0:
                    head_whitespace = top_level[:i]
                top_level = top_level[i:]
                break
        tail_whitespace = ''
        for i in range(len(top_level)):
            if not top_level[-(i+1)].isspace():
                if i > 0:
                    tail_whitespace = top_level[-i:]
                    top_level = top_level[:-i]
                break
        ret.append((head_whitespace, parse_relation(top_level),
                    tail_whitespace))
    return ret


def format_relations(
        relations: List[Tuple[str, List[PkgRelation], str]]) -> str:
    """Format a package relations string.

    This attemps to create formatting.
    """
    ret = []
    for (head_whitespace, relation, tail_whitespace) in relations:
        ret.append(head_whitespace + ' | '.join(o.str() for o in relation) +
                   tail_whitespace)
    # The first line can be whitespace only, the subsequent ones can not
    lines = []
    for i, line in enumerate(','.join(ret).split('\n')):
        if i == 0:
            lines.append(line)
        else:
            if line.strip():
                lines.append(line)
    return '\n'.join(lines)


def get_relation(
        relationstr: str, package: str) -> Tuple[int, List[PkgRelation]]:
    """Retrieve the relation for a particular package.

    Args:
      relationstr: package relation string
      package: package name
    Returns:
      Tuple with offset and relation object
    """
    for offset, relation in iter_relations(relationstr, package):
        names = [r.name for r in relation]
        if len(names) > 1 and package in names:
            raise ValueError("Complex rule for %s , aborting" % package)
        if names != [package]:
            continue
        return offset, relation
    raise KeyError(package)


def iter_relations(
        relationstr: str,
        package: str) -> Iterable[Tuple[int, List[PkgRelation]]]:
    """Iterate over the relations relevant for a particular package.

    Args:
      relationstr: package relation string
      package: package name
    Yields:
      Tuples with offset and relation objects
    """
    relations = parse_relations(relationstr)
    return _iter_relations(relations, package)


def _iter_relations(relations, package):
    for i, (_head_whitespace, relation, _tail_whitespace) in enumerate(
            relations):
        if isinstance(relation, str):  # formatting
            continue
        names = [r.name for r in relation]
        if package not in names:
            continue
        yield i, relation


def ensure_minimum_version(
        relationstr: str, package: str,
        minimum_version: Union[str, Version]) -> str:
    """Update a relation string to ensure a particular version is required.

    Args:
      relationstr: package relation string
      package: package name
      minimum_version: Minimum version
    Returns:
      updated relation string
    """
    def is_obsolete(relation):
        for r in relation:
            if r.name != package:
                continue
            if r.version[0] == '>>' and r.version[1] < minimum_version:
                return True
            if r.version[0] == '>=' and r.version[1] <= minimum_version:
                return True
        return False

    minimum_version = Version(minimum_version)
    found = False
    changed = False
    relations = parse_relations(relationstr)
    obsolete_relations = []
    for i, (_head_whitespace, relation, _tail_whitespace) in enumerate(
            relations):
        if isinstance(relation, str):  # formatting
            continue
        names = [r.name for r in relation]
        if len(names) > 1 and package in names and is_obsolete(relation):
            obsolete_relations.append(i)
        if names != [package]:
            continue
        found = True
        if (relation[0].version is None or
                Version(relation[0].version[1]) < minimum_version):
            relation[0].version = ('>=', minimum_version)
            changed = True
    if not found:
        changed = True
        _add_relation(
            relations,
            [PkgRelation(name=package, version=('>=', minimum_version))])
    for i in reversed(obsolete_relations):
        del relations[i]
    if changed:
        return format_relations(relations)
    # Just return the original; we don't preserve all formatting yet.
    return relationstr


def ensure_exact_version(
        relationstr: str, package: str,
        version: Union[str, Version], position: Optional[int] = None) -> str:
    """Update a relation string to depend on a specific version.

    Args:
      relationstr: package relation string
      package: package name
      version: Exact version to depend on
      position: Optional position in the list to insert any new entries
    Returns:
      updated relation string
    """
    version = Version(version)
    found = False
    changed = False
    relations = parse_relations(relationstr)
    for (_head_whitespace, relation, _tail_whitespace) in relations:
        if isinstance(relation, str):  # formatting
            continue
        names = [r.name for r in relation]
        if len(names) > 1 and names[0] == package:
            raise Exception("Complex rule for %s , aborting" % package)
        if names != [package]:
            continue
        found = True
        if (relation[0].version is None or
                (relation[0].version[0],
                 Version(relation[0].version[1])) != ('=', version)):
            relation[0].version = ('=', version)
            changed = True
    if not found:
        changed = True
        _add_relation(
            relations,
            [PkgRelation(name=package, version=('=', version))],
            position=position)
    if changed:
        return format_relations(relations)
    # Just return the original; we don't preserve all formatting yet.
    return relationstr


def ensure_relation(
        relationstr: str, new_relationstr: Union[str, List[PkgRelation]],
        position: Optional[int] = None) -> str:
    """Ensure that a relation exists.

    This is done by either verifying that there is an existing
    relation that satisfies the specified relation, or
    by upgrading an existing relation.
    """
    if isinstance(new_relationstr, str):
        new_relation = parse_relation(new_relationstr)
    else:
        new_relation = new_relationstr
    relations = parse_relations(relationstr)
    added = False
    to_remove = []
    for i, (_head_whitespace, relation, _tail_whitespace) in enumerate(
            relations):
        if isinstance(relation, str):  # formatting
            continue
        if is_relation_implied(new_relation, relation):
            return relationstr
        if is_relation_implied(relation, new_relation):
            if added:
                to_remove.append(i)
            else:
                relations[i] = (relations[i][0], new_relation, relations[i][2])
                added = True
    if not added:
        _add_relation(relations, new_relation)

    for i in reversed(to_remove):
        del relations[i]

    return format_relations(relations)


def _add_relation(
        relations: List[Tuple[str, List[PkgRelation], str]],
        relation: List[PkgRelation],
        position: Optional[int] = None) -> None:
    """Add a relation to a list of relations.

    Args:
      relations: existing list of relations
      relation: New relation
      position: Optional position to insert the new relation
    Returns:
      Nothing
    """
    if len(relations) > 0 and not relations[-1][1]:
        pointless_tail: Optional[Tuple[str, List[PkgRelation], str]]
        pointless_tail = relations.pop(-1)
    else:
        pointless_tail = None
    if len(relations) == 0:
        head_whitespace = ''
        tail_whitespace = ''
    elif len(relations) == 1:
        head_whitespace = (relations[0][0] or " ")  # Best guess
        tail_whitespace = ''
    else:
        hws: Dict[str, int] = collections.defaultdict(lambda: 0)
        for r in relations[1:]:
            hws[r[0]] += 1
        if len(hws) == 1:
            head_whitespace = list(hws.keys())[0]
        else:
            head_whitespace = relations[-1][0]  # Best guess
        tws: Dict[str, int] = collections.defaultdict(lambda: 0)
        for r in relations[0:-1]:
            tws[r[2]] += 1
        if len(tws) == 1:
            tail_whitespace = list(tws.keys())[0]
        else:
            tail_whitespace = relations[0][2]  # Best guess

    if position is None:
        position = len(relations)

    if position < 0 or position > len(relations):
        raise IndexError('position out of bounds: %r' % position)

    if position == len(relations):
        if len(relations) == 0:
            last_tail_whitespace = ''
        else:
            last_tail_whitespace = relations[-1][2]
            relations[-1] = relations[-1][:2] + (tail_whitespace, )
        relations.append((head_whitespace, relation, last_tail_whitespace))
    elif position == 0:
        relations.insert(
            position, (relations[0][0], relation, tail_whitespace))
        relations[1] = (head_whitespace, relations[1][1], relations[1][2])
    else:
        relations.insert(
            position, (head_whitespace, relation, tail_whitespace))
    if pointless_tail:
        relations.append(pointless_tail)


def add_dependency(relationstr, relation, position=None):
    """Add a dependency to a depends line.

    Args:
      relationstr: existing relations line
      relation: New relation
      position: Optional position to insert relation at (defaults to last)
    Returns:
      Nothing
    """
    relations = parse_relations(relationstr)
    if isinstance(relation, str):
        relation = parse_relation(relation)
    _add_relation(relations, relation, position=position)
    return format_relations(relations)


def ensure_some_version(relationstr: str, package: str) -> str:
    """Add a package dependency to a depends line if it's not there.

    Args:
      relationstr: existing relations line
      package: Package to add dependency on
    Returns:
      new formatted relation string
    """
    if not isinstance(package, str):
        raise TypeError(package)
    relations = parse_relations(relationstr)
    for (_head_whitespace, relation, _tail_whitespace) in relations:
        if isinstance(relation, str):  # formatting
            continue
        names = [r.name for r in relation]
        if len(names) > 1 and names[0] == package:
            raise Exception("Complex rule for %s , aborting" % package)
        if names != [package]:
            continue
        return relationstr
    _add_relation(relations, parse_relation(package))
    return format_relations(relations)


def filter_dependencies(
        relations: List[Tuple[str, List[PkgRelation], str]],
        keep: Callable[[List[PkgRelation]], bool]
        ) -> List[Tuple[str, List[PkgRelation], str]]:
    """Filter out some dependencies.

    Args:
      relations: List of relations
      keep: callback that receives a relation and returns a boolean
    Returns:
      Updated list of relations
    """
    ret = []
    for i, entry in enumerate(relations):
        (head_whitespace, relation, tail_whitespace) = entry
        if isinstance(relation, str):  # formatting
            ret.append(entry)
            continue
        if keep(relation):
            ret.append(entry)
            continue
        elif i == 0 and len(relations) > 1:
            # If the first item is removed, then copy the spacing to the next
            # item
            relations[1] = (head_whitespace, relations[1][1], tail_whitespace)
    return ret


def drop_dependency(relationstr: str, package: str) -> str:
    """Drop a dependency from a depends line.

    Args:
      relationstr: package relation string
      package: package name
    Returns:
      updated relation string
    """
    relations = parse_relations(relationstr)

    def keep(relation):
        names = [r.name for r in relation]
        return set(names) != {package}
    ret = filter_dependencies(relations, keep)
    if relations != ret:
        return format_relations(ret)
    # Just return the original; we don't preserve all formatting yet.
    return relationstr


def delete_from_list(
        liststr: str, item_to_delete: Union[str, List[str]]) -> str:
    if not item_to_delete:
        raise ValueError(item_to_delete)
    if isinstance(item_to_delete, str):
        items_to_delete = [item_to_delete.strip()]
    elif isinstance(item_to_delete, list):
        items_to_delete = [item.strip() for item in item_to_delete]
    else:
        raise TypeError(item_to_delete)
    items = liststr.split(',')
    for i, item in enumerate(items):
        if item.strip() in items_to_delete:
            deleted_item = items.pop(i)
            head_whitespace = ''.join(
                takewhile(lambda x: x.isspace(), deleted_item))
            if i == 0 and len(items) >= 1:
                # If we're removing the first item, copy its whitespace to the
                # second
                items[i] = head_whitespace + items[i].lstrip()
            elif i == len(items):
                if i > 1:
                    items[i-1] = items[i-1].rstrip()
    return ','.join(items)


def is_dep_implied(dep: PkgRelation, outer: PkgRelation) -> bool:
    """Check if one dependency is implied by another.

    Is dep implied by outer?
    """
    if dep.name != outer.name:
        return False
    if not dep.version:
        return True
    if outer.version == dep.version:
        return True
    if not outer.version:
        return False
    if dep.version[0] == '>=':
        if outer.version[0] == '>>':
            return Version(outer.version[1]) > Version(dep.version[1])
        elif outer.version[0] in ('>=', '='):
            return Version(outer.version[1]) >= Version(dep.version[1])
        elif outer.version[0] in ('<<', '<='):
            return False
        else:
            raise AssertionError('unsupported: %s' % outer.version[0])
    elif dep.version[0] == '=':
        if outer.version[0] == '=':
            return Version(outer.version[1]) == Version(dep.version[1])
        else:
            return False
    elif dep.version[0] == '<<':
        if outer.version[0] == '<<':
            return Version(outer.version[1]) <= Version(dep.version[1])
        if outer.version[0] in ('<=', '='):
            return Version(outer.version[1]) < Version(dep.version[1])
        elif outer.version[0] in ('>>', '>='):
            return False
        else:
            raise AssertionError('unsupported: %s' % outer.version[0])
    elif dep.version[0] == '<=':
        if outer.version[0] in ('<=', '=', '<<'):
            return Version(outer.version[1]) <= Version(dep.version[1])
        elif outer.version[0] in ('>>', '>='):
            return False
        else:
            raise AssertionError('unsupported: %s' % outer.version[0])
    elif dep.version[0] == '>>':
        if outer.version[0] == '>>':
            return Version(outer.version[1]) >= Version(dep.version[1])
        elif outer.version[0] in ('=', '>='):
            return Version(outer.version[1]) > Version(dep.version[1])
        elif outer.version[0] in ('<<', '<='):
            return False
        else:
            raise AssertionError('unsupported: %s' % outer.version[0])
    else:
        raise AssertionError('unable to handle checks for %s' % dep.version[0])


def is_relation_implied(
        inner: Union[str, List[PkgRelation]],
        outer: Union[str, List[PkgRelation]]) -> bool:
    """Check if one relation implies another.

    Args:
      inner: Inner relation
      outer: Outer relation
    Return: boolean
    """
    if isinstance(inner, str):
        inner_rel = parse_relation(inner)
    else:
        inner_rel = inner
    if isinstance(outer, str):
        outer_rel = parse_relation(outer)
    else:
        outer_rel = outer

    if inner_rel == outer_rel:
        return True

    # "bzr >= 1.3" implied by "bzr >= 1.3 | libc6"
    for inner_dep in inner_rel:
        if all(is_dep_implied(inner_dep, outer_dep)
               for outer_dep in outer_rel):
            return True
    return False


def parse_standards_version(v: str) -> Tuple[int, ...]:
    """Parse a standards version.

    Args:
      v: Version string
    Returns: Tuple with version
    """
    return tuple([int(k) for k in v.split('.')])


def suppress_substvar_warnings():
    import warnings
    warnings.filterwarnings(
        action='ignore',
        category=UserWarning,
        message=(r'cannot parse package relationship \"\$\{.*\}\", returning '
                 r'it raw'))


class PkgRelationFieldEditor:
    """Convenience wrapper for editing pkg relation fields."""

    _parsed: Optional[List[Tuple[str, List[PkgRelation], str]]]

    def __init__(self, paragraph, name):
        self.paragraph = paragraph
        self.name = name
        self._parsed = None

    def __enter__(self):
        v = self.paragraph.get(self.name)
        if v is not None:
            self._parsed = parse_relations(v)
        else:
            self._parsed = None

        return self

    def drop_relation(self, package):
        """Drop a relation.

        Args:
          package: package name
        """
        if self._parsed is None:
            return

        def keep(relation):
            names = [r.name for r in relation]
            return set(names) != {package}
        new_parsed = filter_dependencies(self._parsed, keep)
        ret = self._parsed != new_parsed
        self._parsed = new_parsed
        return ret

    def __bool__(self):
        return self._parsed is not None

    def add_relation(self, relation, position=None):
        """Add a relation.

        Args:
          relation: New relation
          position: Optional position to insert relation at (defaults to last)
        """
        if isinstance(relation, str):
            relation = parse_relation(relation)
        if not self._parsed:
            self._parsed = []
        return _add_relation(self._parsed, relation, position=position)

    def iter_relations(self, package):
        """Iterate over all relations with a particular package.

        Args:
          package: Package name
        Yields:
          Tuples with offset and relation objects
        """
        if self._parsed is None:
            return
        return _iter_relations(self._parsed, package)

    def get_relation(self, package):
        """Retrieve the relation for a particular package.

        Args:
          relationstr: package relation string
          package: package name
        Raises:
          ValueError if there is more than one relation with the package
        Returns:
          Tuple with offset and relation object
        """
        for offset, relation in self.iter_relations(package):
            names = [r.name for r in relation]
            if len(names) > 1 and package in names:
                raise ValueError("Complex rule for %s , aborting" % package)
            if names != [package]:
                continue
            return offset, relation
        raise KeyError(package)

    def has_relation(self, package):
        """Check whether there is a relation with specified package.

        Args:
          package: name of the package
        Returns: boolean
        """
        try:
            self.get_relation(package)
        except KeyError:
            return False
        else:
            return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            if self._parsed is None:
                try:
                    del self.paragraph[self.name]
                except KeyError:
                    pass
            else:
                self.paragraph[self.name] = format_relations(self._parsed)
        return False


def format_description(summary, long_description):
    """Format a description based on summary and long description lines.
    """
    return summary + "\n" + ''.join(
        [" %s\n" % line for line in long_description])



#!/usr/bin/python3
# Copyright (C) 2019-2020 Jelmer Vernooij
#
# find_extra_authors and find_thanks originally imported from
# breezy-debian and (C) 2006 James Westby
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with changelog files."""

__all__ = [
    'ChangelogParseError',
    'ChangelogCreateError',
    'ChangelogEditor',
    'changes_sections',
    'changes_by_author',
    'changelog_add_entry',
    'changeblock_add_line',
    'changeblock_ensure_first_line',
    'all_sha_prefixed',
    'any_long_lines',
    'find_extra_authors',
    'find_thanks',
    'rewrap_change',
    'strip_changelog_message',
    'new_changelog_entries',
    'is_unreleased_inaugural',
    'upstream_merge_changelog_line',
    ]

from datetime import datetime
from email.utils import format_datetime, parseaddr
import re
import textwrap
from typing import List, Iterator, Tuple, Optional

from debian.changelog import (
    ChangeBlock,
    Changelog,
    ChangelogCreateError,
    ChangelogParseError,
    format_date,
    get_maintainer,
    Version,
    )
from .reformatting import Editor

WIDTH = 80
INITIAL_INDENT = '  * '
DEFAULT_DISTRIBUTION = 'unstable'


class ChangelogEditor(Editor[Changelog, str]):
    """Update a changelog file.

    This will only write out the changelog file if it has been changed.
    """

    def __init__(
            self, path: str = 'debian/changelog',
            allow_reformatting: Optional[bool] = False,
            allow_missing: bool = False):
        super().__init__(
            path, allow_reformatting=allow_reformatting)
        self.allow_missing = allow_missing

    @classmethod
    def create(cls, path: str = 'debian/changelog'):
        return cls(path, allow_reformatting=True, allow_missing=True)

    def _parse(self, content):
        cl = Changelog()
        cl.parse_changelog(
            content, max_blocks=None, allow_empty_author=True, strict=False)
        return cl

    def _format(self, parsed):
        return parsed._format(allow_missing_author=True)

    @property
    def changelog(self):
        return self._parsed

    def __getitem__(self, i):
        return self.changelog[i]

    def _nonexistant(self):
        if self.allow_missing:
            return Changelog()
        raise

    def new_block(self, *args, **kwargs):
        return self.changelog.new_block(*args, **kwargs)

    def auto_version(self, version, **kwargs):
        return changelog_auto_version(
            self.changelog, version=version, **kwargs)

    def add_entry(
            self,
            summary: List[str],
            maintainer: Optional[Tuple[str, str]] = None,
            timestamp: Optional[datetime] = None,
            urgency: str = 'low') -> None:
        return changelog_add_entry(
            self.changelog, summary=summary, maintainer=maintainer,
            timestamp=timestamp, urgency=urgency)


def changelog_auto_version(
        cl: Changelog, version: Version,
        maintainer: Optional[Tuple[str, str]] = None,
        timestamp: Optional[datetime] = None,
        package: Optional[str] = None,
        urgency: str = 'low'):
    """Update current changelog entry to version or create a new one.
    """
    if maintainer is None:
        maintainer_name, maintainer_email = get_maintainer()
    else:
        maintainer_name, maintainer_email = maintainer
    if maintainer_name is None or maintainer_email is None:
        raise ValueError(
            'unable to determine maintainer details and none specified')
    if package is None:
        package = cl[0].package
    if timestamp is None:
        timestamp = datetime.now()
    if len(cl) > 0 and distribution_is_unreleased(cl[0].distributions):
        cl[0].version = version
        cl[0].date = format_datetime(timestamp)
        cl[0].package = package
        # TODO(jelmer): Also set maintainer again?
    else:
        cl.new_block(
            version=version, package=package,
            distributions='UNRELEASED', urgency=urgency,
            author="{} <{}>".format(maintainer_name, maintainer_email),
            date=format_datetime(timestamp))


def changes_sections(
        changes: List[str]
        ) -> Iterator[
            Tuple[Optional[str], List[int], List[List[Tuple[int, str]]]]
        ]:
    """Return the different sections from a set of changelog entries.

    Args:
      changes: list of changes from a changelog entry
    Returns:
      iterator over tuples with:
        (author, list of line numbers, list of list of (lineno, line) tuples
    """
    section: Tuple[
        Optional[str], List[int],
        List[List[Tuple[int, str]]]] = (None, [], [])
    change: List[Tuple[int, str]] = []
    for i, line in enumerate(changes):
        if not line and i == 0:
            # Skip the first line
            continue
        if not line:
            section[1].append(i)
            continue
        m = re.fullmatch(r'  \[ (.*) \]', line)
        if m:
            if change:
                section[2].append(change)
                change = []
            if section[1]:
                yield section
            section = (m.group(1), [i], [])
            continue
        if not line.startswith(INITIAL_INDENT):
            change.append((i, line))
            section[1].append(i)
        else:
            if change:
                section[2].append(change)
            change = [(i, line)]
            section[1].append(i)
    if change:
        section[2].append(change)
    if section[1]:
        yield section


def changes_by_author(
        changes: List[str]
        ) -> Iterator[Tuple[Optional[str], List[int], List[str]]]:
    """Changes by author.

    Args:
      changes: Find changes by author from a list of lines
    Returns:
      Iterator over items by author (maintainer, offsets, changes)
    """
    for (author, _linenos, contents) in changes_sections(changes):
        for change_entries in contents:
            change_linenos, change_lines = zip(*change_entries)
            yield (author, change_linenos, change_lines)  # type: ignore


class TextWrapper(textwrap.TextWrapper):

    whitespace = r'[%s]' % re.escape('\t\n\x0b\x0c\r ')
    wordsep_simple_re = re.compile(r'(%s+)' % whitespace)

    def __init__(self, initial_indent=INITIAL_INDENT):
        super().__init__(
            width=WIDTH, initial_indent=initial_indent,
            subsequent_indent=' ' * len(initial_indent),
            break_long_words=False, break_on_hyphens=False)

    def _split(self, text):
        chunks = [c for c in self.wordsep_simple_re.split(text) if c]
        ret = []
        i = 0
        while i < len(chunks):
            if (any(chunks[i].endswith(x) for x in ['Closes:', 'LP:']) and
                    i+2 < len(chunks) and chunks[i+2].startswith('#')):
                ret.append('{} {}'.format(chunks[i], chunks[i+2]))
                i += 3
            else:
                ret.append(chunks[i])
                i += 1
        return ret


_initial_re = re.compile(r'^[  ]+[\+\-\*] ')


def _can_join(line1, line2):
    if line1.endswith(':'):
        return False
    if line2 and line2[:1].isupper():
        if line1.endswith(']') or line1.endswith('}'):
            return False
        if not line1.endswith('.'):
            return False
    return True


def any_long_lines(lines: List[str], width: int = WIDTH) -> bool:
    """Check if any lines are longer than the specified width.
    """
    return any([len(line) > width for line in lines])


def rewrap_change(change: List[str]) -> List[str]:
    """Rewrap lines from a list of changes.

    Args:
      change: List of lines
    Returns:
      new list of lines
    """
    if not change:
        return change
    m = _initial_re.match(change[0])
    if not any_long_lines(change) or not m:
        return change
    wrapper = TextWrapper(m.group(0))
    prefix_len = len(m.group(0))
    lines = [line[prefix_len:] for line in change]
    todo = [lines[0]]
    ret = []
    for i in range(len(lines) - 1):
        if _can_join(lines[i], lines[i+1]) and any_long_lines(
                todo, WIDTH-prefix_len):
            todo.append(lines[i+1])
        else:
            ret.extend(wrapper.wrap('\n'.join(todo)))
            wrapper = TextWrapper(change[i+1][:len(m.group(0))])
            todo = [lines[i+1]]
    ret.extend(wrapper.wrap('\n'.join(todo)))
    return ret


def rewrap_changes(changes: Iterator[str]) -> Iterator[str]:
    change: List[str] = []
    indent = None
    for line in changes:
        m = _initial_re.match(line)
        if m:
            yield from rewrap_change(change)
            change = [line]
            indent = len(m.group(0))
        elif change and line.startswith(' ' * indent):  # type: ignore
            change.append(line)
        else:
            yield from rewrap_change(change)
            change = []
            yield line


def increment_version(version: Version) -> Version:
    """Increment a Debian version string."""
    ret = Version(str(version))
    # TODO(jelmer): Add ubuntuX suffix on Ubuntu
    if ret.debian_revision:
        # Non-native package
        m = re.match('^(.*?)([0-9]+)$', ret.debian_revision)
        if m:
            ret.debian_revision = m.group(1) + str(int(m.group(2))+1)
        else:
            ret.debian_revision += "1"
        return ret
    elif ret.upstream_version:
        # Native package
        m = re.match('^(.*?)([0-9]+)$', ret.upstream_version)
        if m:
            ret.upstream_version = m.group(1) + str(int(m.group(2))+1)
        else:
            ret.upstream_version += "1"
        return ret
    else:
        # Uhm..
        raise ValueError(ret)


def changelog_add_entry(
        cl: Changelog, summary: List[str],
        maintainer: Optional[Tuple[str, str]] = None,
        timestamp: Optional[datetime] = None,
        urgency: str = 'low') -> None:
    """Add an entry to a changelog.

    Args:
      cl: Changelog object to modify
      summary: List of lines
      maintainer: maintainer identity to use (defaults to get_maintainer())
      timestamp: Timestamp to set for new entries
      urgency: Urgency to use for new items
    Raises:
      ValueError: when the maintainer details can not be found and were not
        specified
    """
    if timestamp is None:
        timestamp = datetime.now()
    if maintainer is None:
        maintainer_name, maintainer_email = get_maintainer()
    else:
        maintainer_name, maintainer_email = maintainer
    if maintainer_name is None or maintainer_email is None:
        raise ValueError(
            'unable to determine maintainer details and none specified')
    if (distribution_is_unreleased(cl[0].distributions) or (
                cl[0].author is None and cl[0].date is None)):
        by_author = list(changes_by_author(cl[0].changes()))
        if cl[0]._changes == []:
            cl[0]._changes.append('')
        if all([author is None for (author, linenos, change) in by_author]):
            if cl[0].author is not None:
                entry_maintainer = parseaddr(cl[0].author)
                if entry_maintainer != (maintainer_name, maintainer_email):
                    cl[0]._changes.insert(1, '  [ %s ]' % entry_maintainer[0])
                    if cl[0]._changes[-1]:
                        cl[0]._changes.append('')
                    cl[0]._changes.append('  [ %s ]' % maintainer_name)
        else:
            if by_author[-1][0] != maintainer_name:
                if cl[0]._changes[-1]:
                    cl[0]._changes.append('')
                cl[0]._changes.append('  [ %s ]' % maintainer_name)
        if len(cl[0] ._changes) > 1 and not cl[0]._changes[-1].strip():
            del cl[0]._changes[-1]
    else:
        cl.new_block(
            package=cl[0].package,
            version=increment_version(cl[0].version),
            urgency=urgency,
            author="{} <{}>".format(maintainer_name, maintainer_email),
            date=format_datetime(timestamp),
            distributions='UNRELEASED',
            changes=[''])
    changeblock_add_line(cl[0], summary)


def changeblock_add_line(block, lines):
    wrapper = TextWrapper(INITIAL_INDENT)
    block._changes.extend(wrapper.wrap(lines[0]))
    for line in lines[1:]:
        prefix = len(INITIAL_INDENT) * ' '
        m = re.match(r'^[  ]*[\+\-\*] ', line)
        if m:
            prefix += m.group(0)
            line = line[len(m.group(0)):]
        block._changes.extend(TextWrapper(prefix).wrap(line))
    block._changes.append('')


def strip_changelog_message(changes: List[str]) -> List[str]:
    """Strip a changelog message like debcommit does.

    Takes a list of changes from a changelog entry and applies a transformation
    so the message is well formatted for a commit message.

    :param changes: a list of lines from the changelog entry
    :return: another list of lines with blank lines stripped from the start
        and the spaces the start of the lines split if there is only one
        logical entry.
    """
    if not changes:
        return changes
    while changes and changes[-1] == '':
        changes.pop()
    while changes and changes[0] == '':
        changes.pop(0)

    whitespace_column_re = re.compile(r'  |\t')
    changes = [whitespace_column_re.sub('', line, 1) for line in changes]

    leader_re = re.compile(r'[ \t]*[*+-] ')
    count = len([line for line in changes if leader_re.match(line)])
    if count == 1:
        return [leader_re.sub('', line, 1).lstrip() for line in changes]
    else:
        return changes


def new_changelog_entries(
        old_text: List[bytes], new_text: List[bytes]) -> List[str]:
    import difflib
    sequencematcher = difflib.SequenceMatcher
    changes = []
    for group in sequencematcher(
            None, old_text, new_text).get_grouped_opcodes(0):
        j1, j2 = group[0][3], group[-1][4]
        for line in new_text[j1:j2]:
            if line.startswith(b"  "):
                # Debian Policy Manual states that debian/changelog must be
                # UTF-8
                changes.append(line.decode('utf-8'))
    return changes


def new_upstream_package_version(
        upstream_version: str, distribution_name: str,
        epoch: Optional[str] = None) -> Version:
    """Determine the package version for a new upstream.

    Args:
      upstream_version: Upstream version string
      distribution_name: Distribution the package is for
      epoch: Optional epoch
    Returns:
      The new version
    """
    if distribution_name == "ubuntu":
        ret = Version("%s-0ubuntu1" % upstream_version)
    else:
        ret = Version("%s-1" % upstream_version)
    ret.epoch = epoch
    return ret


def all_sha_prefixed(cb: ChangeBlock) -> bool:
    """Check if all lines in a changelog entry are prefixed with a sha.

    This is generally done by gbp-dch(1).

    Args:
      cl: Changelog entry
    Returns: bool
    """
    sha_prefixed = 0
    for change in cb.changes():
        if not change.startswith('  * '):
            continue
        if re.match(r'  \* \[[0-9a-f]{7}\] ', change):
            sha_prefixed += 1
        else:
            return False
    return (sha_prefixed > 0)


def distribution_is_unreleased(distribution):
    return (
        distribution == 'UNRELEASED' or
        distribution.startswith('UNRELEASED-'))


def changeblock_ensure_first_line(
        block, line, maintainer: Optional[Tuple[str, str]] = None):
    """Ensure that the first line matches the specified line.

    Args:
      block: Changelog block to edit
      line: Line to add
      maintainer: Maintainer to reference as (name, email)-tuple
    """
    if maintainer is None:
        maintainer_name, maintainer_email = get_maintainer()
    else:
        maintainer_name, maintainer_email = maintainer
    if maintainer_name is None or maintainer_email is None:
        raise ValueError(
            'unable to determine maintainer details and none specified')
    if block._changes[0]:
        raise ValueError('first block line not empty')
    line = '  * ' + line
    if block._changes[1] == line:
        return
    block._changes.insert(1, line)
    if block._changes[2].startswith('  ['):
        block._changes.insert(2, '')
    elif parseaddr(block.author)[0] != maintainer_name:
        block._changes.insert(2, '  [ %s ]' % parseaddr(block.author)[0])
        block._changes.insert(2, '')
        block.author = "{} <{}>".format(maintainer_name, maintainer_email)


def take_uploadership(
        block, maintainer: Optional[Tuple[str, str]] = None) -> None:
    """Take uploaderhsip of a changelog entry, but attribute contributors.

    Args:
      block: Changelog block to modify
      maintainer: Tuple with (name, email) of maintainer to take ownership
    """
    if maintainer is None:
        maintainer_name, maintainer_email = get_maintainer()
    else:
        maintainer_name, maintainer_email = maintainer
    if maintainer_name is None or maintainer_email is None:
        raise ValueError(
            'unable to determine maintainer details and none specified')
    if block.author is not None:
        entry_maintainer = parseaddr(block.author)
        if (entry_maintainer != (maintainer_name, maintainer_email) and
                len(block._changes) >= 2 and
                not block._changes[1].startswith('  [ ')):
            block._changes.insert(1, '  [ %s ]' % entry_maintainer[0])
            if block._changes[-1]:
                block._changes.append('')
    block.author = '{} <{}>'.format(maintainer_name, maintainer_email)


def release(
        cl, distribution: Optional[str] = None, timestamp=None,
        localtime: bool = True,
        maintainer: Optional[Tuple[str, str]] = None):
    """Create a release for a changelog file.
    """
    if distribution is None:
        try:
            distribution = cl[1].distributions
        except IndexError:
            distribution = DEFAULT_DISTRIBUTION
    if distribution_is_unreleased(cl[0].distributions):
        take_uploadership(cl[0], maintainer)
        cl[0].distributions = distribution
        cl[0].date = format_date(timestamp=timestamp, localtime=localtime)


def find_last_distribution(changelog):
    """Find the last changelog that was used in a changelog.

    This will skip stanzas with the 'UNRELEASED' distribution.

    Args:
      changelog: Changelog to analyze
    """
    for block in changelog._blocks:
        distribution = block.distributions.split(" ")[0]
        if not distribution_is_unreleased(distribution):
            return distribution
    return None


def find_extra_authors(changes):
    """Find additional authors from a changelog entry.

    :return: List of fullnames of additional authors, without e-mail address.
    """
    authors: List[str] = []
    for new_author, _linenos, _lines in changes_by_author(changes):
        if new_author is None:
            continue
        already_included = False
        for author in authors:
            if author.startswith(new_author):
                already_included = True
                break
        if not already_included:
            authors.append(new_author)
    return authors


def find_thanks(changes):
    """Find all people thanked in a changelog entry.

    :param changes: String with the contents of the changelog entry
    :return: List of people thanked, optionally including email address.
    """
    thanks_re = re.compile(
        r"[tT]hank(?:(?:s)|(?:you))(?:\s*to)?"
        "((?:\\s+(?:(?:\\w\\.)|(?:\\w+(?:-\\w+)*)))+"
        "(?:\\s+<[^@>]+@[^@>]+>)?)",
        re.UNICODE)
    thanks: List[str] = []
    for _new_author, _linenos, lines in changes_by_author(changes):
        for match in thanks_re.finditer(''.join(lines)):
            if thanks is None:
                thanks = []
            thanks_str = match.group(1).strip()
            thanks_str = re.sub(r"\s+", " ", thanks_str)
            thanks.append(thanks_str)
    return thanks


def upstream_merge_changelog_line(upstream_version: str) -> str:
    """Describe that a new upstream revision was merged.

    This will either describe that a new upstream release or a new upstream
    snapshot was merged.

    Args:
      upstream_version: Upstream version string
    Returns:
       line string for use in changelog
    """
    vcs_suffixes = ["~bzr", "+bzr", "~svn", "+svn", "~git", "+git", "-git"]
    for vcs_suffix in vcs_suffixes:
        if vcs_suffix in str(upstream_version):
            entry_description = "New upstream snapshot."
            break
    else:
        entry_description = "New upstream release."
    return entry_description


def is_unreleased_inaugural(cl: Changelog) -> bool:
    """Check whether this is a traditional inaugural release.

    Args:
      cl: A changelog object to inspect
    """
    if cl is None:
        return False
    if len(cl) != 1:
        return False
    if not distribution_is_unreleased(cl[0].distributions):
        return False
    actual = [change for change in cl[0].changes() if change.strip()]
    if len(actual) != 1:
        return False
    if not actual[0].startswith('  * Initial release'):
        return False
    return True


def gbp_dch(path: str) -> None:
    """Run 'gbp dch'."""
    import os
    from gbp.scripts.dch import main as dch_main
    old_cwd = os.getcwd()
    try:
        os.chdir(path)
        dch_main(['gbp-dch', '--ignore-branch'])
    finally:
        os.chdir(old_cwd)



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Functions for working with watch files."""

from io import StringIO
import logging
import pcre
import re
import sys
from typing import (
    Iterable, List, Union, Callable, Optional, TextIO, Iterator, Tuple)
from urllib.parse import urljoin

from debian.changelog import Version

from .reformatting import (
    Editor,
    )

from . import __version__

DEFAULT_USER_AGENT = 'debmutate/%s' % '.'.join([str(x) for x in __version__])

DEFAULT_VERSION: int = 4

SUBSTITUTIONS = {
    # This is substituted with the source package name found in the first line
    # of the debian/changelog file.
    # '@PACKAGE@': None,
    # This is substituted by the legal upstream version regex (capturing).
    '@ANY_VERSION@': r'[-_]?(\d[\-+\.:\~\da-zA-Z]*)',
    # This is substituted by the typical archive file extension regex
    # (non-capturing).
    '@ARCHIVE_EXT@': r'(?i)\.(?:tar\.xz|tar\.bz2|tar\.gz|zip|tgz|tbz|txz)',
    # This is substituted by the typical signature file extension regex
    # (non-capturing).
    '@SIGNATURE_EXT@':
        r'(?i)\.(?:tar\.xz|tar\.bz2|tar\.gz|zip|tgz|tbz|txz)'
        r'\.(?:asc|pgp|gpg|sig|sign)',
    # This is substituted by the typical Debian extension regexp (capturing).
    '@DEB_EXT@': r'[\+~](debian|dfsg|ds|deb)(\.)?(\d+)?$',
}


class InvalidUVersionMangle(ValueError):
    """uversionmangle is invalid"""


class WatchFile:

    def __init__(self, entries: Optional[List['Watch']] = None,
                 options: Optional[List[str]] = None,
                 version: int = DEFAULT_VERSION) -> None:
        self.version = version
        if entries is None:
            entries = []
        self.entries = entries
        if options is None:
            options = []
        self.options = options

    def get_option(self, name):
        for option in self.options:
            try:
                key, value = option.split('=', 1)
            except ValueError:
                key = option
                value = None
            if key == name:
                return value
        raise KeyError(name)

    def set_option(self, name, newvalue=None):
        if newvalue is None:
            nv = name
        else:
            nv = '{}={}'.format(name, newvalue)
        for i, option in enumerate(self.options):
            try:
                key, value = option.split('=', 1)
            except ValueError:
                key = option
                value = None
            if key == value:
                self.options[i] = nv
                return
        self.options.append(nv)

    def del_option(self, name):
        for i, option in enumerate(self.options):
            try:
                key, value = option.split('=', 1)
            except ValueError:
                key = option
            if key == name:
                del self.options[i]
                return
        raise KeyError(name)

    def __iter__(self) -> Iterator['Watch']:
        return iter(self.entries)

    def __bool__(self) -> bool:
        return bool(self.entries)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and \
                self.entries == other.entries and \
                self.options == other.options and \
                self.version == other.version

    def dump(self, f: TextIO) -> None:
        def serialize_options(opts: List[str]) -> str:
            s = ','.join(opts)
            if ' ' in s or '\t' in s:
                return 'opts="' + s + '"'
            return 'opts=' + s
        f.write('version=%d\n' % self.version)
        if self.options:
            f.write(serialize_options(self.options) + '\n')
        for entry in self.entries:
            if entry.options:
                f.write(serialize_options(entry.options) + ' ')
            f.write(entry.url)
            if entry.matching_pattern:
                f.write(' ' + entry.matching_pattern)
            if entry.version:
                f.write(' ' + entry.version)
            if entry.script:
                f.write(' ' + entry.script)
            f.write('\n')


def parse_sed_expr(vm):
    if vm.startswith('s'):
        return ('s', parse_subst_expr(vm))
    if vm.startswith('tr'):
        return ('tr', parse_transl_expr(vm))
    if vm.startswith('y'):
        return ('y', parse_transl_expr(vm))
    raise InvalidUVersionMangle(vm, 'not a substitution or translation regex')


def parse_subst_expr(vm: str) -> Tuple[str, str, Optional[str]]:
    if vm[0] != 's':
        raise InvalidUVersionMangle(vm, 'not a substitution regex')
    parts = pcre.split(r'(?<!\\)' + vm[1], vm)
    if len(parts) < 3:
        raise InvalidUVersionMangle(vm)
    pattern = parts[1]
    replacement = parts[2]
    flags: Optional[str]
    try:
        flags = parts[3]
    except IndexError:
        flags = None
    return (pattern, replacement, flags)


def parse_transl_expr(vm: str) -> Tuple[str, str, Optional[str]]:
    if vm.startswith('tr'):
        s = vm[2:]
    elif vm.startswith('y'):
        s = vm[1:]
    else:
        raise InvalidUVersionMangle(vm, 'not a translation regex')
    parts = pcre.split(r'(?<!\\)' + s[0], vm)
    if len(parts) < 3:
        raise InvalidUVersionMangle(vm)
    pattern = parts[1]
    replacement = parts[2]
    flags: Optional[str]
    try:
        flags = parts[3]
    except IndexError:
        flags = None
    return (pattern, replacement, flags)


def apply_sed_expr(vm: str, orig: str) -> str:
    (kind, (pattern, replacement, flags)) = parse_sed_expr(vm)
    if kind == 's':
        # TODO(jelmer): Handle flags
        replacement = re.sub(
            r'\$([0-9]+)', lambda x: '{%s}' % x.group(1), replacement)
        return pcre.sub(pattern, replacement, orig)
    elif kind == 'tr':
        from tr import tr
        return tr(pattern, replacement, orig, flags or '')
    elif kind == 'y':
        from tr import tr
        return tr(pattern, replacement, orig, flags or '')
    else:
        raise ValueError(kind)


def apply_url_mangle(expr: str, orig: str) -> str:
    return apply_sed_expr(expr, orig)


class Release:
    """A discovered release."""

    def __init__(self, version, url, pgpsigurl=None):
        self.version = version
        self.url = url
        self.pgpsigurl = pgpsigurl

    def __lt__(self, other):
        if type(self) != type(other):
            raise TypeError(other)
        return Version(self.version) < Version(other.version)

    def __repr__(self):
        return "{}({!r}, {!r}, pgpsigurl={!r})".format(
            type(self).__name__, self.version, self.url,
            self.pgpsigurl)


def html_search(body, matching_pattern, base_url):
    from bs4 import BeautifulSoup
    soup = BeautifulSoup(body, 'html.parser')
    if '/' not in matching_pattern:
        matching_pattern = urljoin(
            base_url.rstrip('/') + '/', matching_pattern)
    for a in soup.find_all('a'):
        href = a.attrs.get('href')
        if not href:
            continue
        href = urljoin(base_url.rstrip('/') + '/', href)
        m = pcre.match(matching_pattern, href)
        if m:
            logging.debug('Matched pattern %r to %r', matching_pattern, href)
            yield m
        else:
            logging.debug(
                'Did not match pattern %r to %r', matching_pattern, href)


def plain_search(body, matching_pattern, base_url):
    return pcre.finditer(matching_pattern.encode(), body)


searchers = {
    'plain': plain_search,
    'html': html_search,
    }


def _subst(text: str, package: Union[str, Callable[[], str]]):
    substs = dict(SUBSTITUTIONS)
    if '@PACKAGE@' in text:
        if callable(package):
            package = package()
        substs['@PACKAGE@'] = package
    for k, v in substs.items():
        text = text.replace(k, v)
    return text


class Watch:

    def __init__(self, url: str, matching_pattern: Optional[str] = None,
                 version: Optional[str] = None, script: Optional[str] = None,
                 opts: Optional[List[str]] = None) -> None:
        self.url = url
        self.matching_pattern = matching_pattern
        self.version = version
        self.script = script
        if opts is None:
            opts = []
        self.options = opts

    def uversionmangle(self, version):
        try:
            vm = self.get_option('uversionmangle')
        except KeyError:
            return version
        try:
            return apply_sed_expr(vm, version)
        except pcre.error as e:
            raise WatchSyntaxError(
                'invalid uversionmangle {!r}: {}'.format(vm, e)) from e

    def get_option(self, name):
        for option in self.options:
            try:
                key, value = option.split('=', 1)
            except ValueError:
                key = option
                value = None
            if key == name:
                return value
        raise KeyError(name)

    def has_option(self, name):
        try:
            self.get_option(name)
        except KeyError:
            return False
        return True

    def set_option(self, name, newvalue=None):
        if newvalue is None:
            nv = name
        else:
            nv = '{}={}'.format(name, newvalue)
        for i, option in enumerate(self.options):
            try:
                key, value = option.split('=', 1)
            except ValueError:
                key = option
                value = None
            if key == value:
                self.options[i] = nv
                return
        self.options.append(nv)

    def del_option(self, name):
        for i, option in enumerate(self.options):
            try:
                key, value = option.split('=', 1)
            except ValueError:
                key = option
            if key == name:
                del self.options[i]
                return
        raise KeyError(name)

    def __repr__(self) -> str:
        return (
            ("{}({!r}, matching_pattern={!r}, "
             "version={!r}, script={!r}, opts={!r})")
            .format(
                self.__class__.__name__, self.url, self.matching_pattern,
                self.version, self.script, self.options))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Watch):
            return False
        return (other.url == self.url and
                other.matching_pattern == self.matching_pattern and
                other.version == self.version and
                other.script == self.script and
                other.options == self.options)

    def format_url(self, package: Union[str, Callable[[], str]]) -> str:
        return _subst(self.url, package)

    def discover(self, package) -> Iterator[Release]:
        from urllib.request import urlopen, Request
        url = self.format_url(package)
        try:
            user_agent = self.get_option('user-agent')
        except KeyError:
            user_agent = DEFAULT_USER_AGENT
        try:
            searchmode = self.get_option('searchmode')
        except KeyError:
            searchmode = 'html'
        logging.debug('Fetching url %s; searchmode=%s', url, searchmode)
        req = Request(url, headers={'User-Agent': user_agent})
        resp = urlopen(req)
        assert self.matching_pattern
        for m in searchers[searchmode](
                resp.read(), _subst(self.matching_pattern, package), url):
            # TODO(jelmer): Apply uversionmangle
            full_url = urljoin(url, m.group(0))
            try:
                pgpsigurlmangle = self.get_option('pgpsigurlmangle')
            except KeyError:
                pgpsigurl = None
            else:
                pgpsigurl = apply_url_mangle(pgpsigurlmangle, full_url)
            yield Release(m.group(1), full_url, pgpsigurl=pgpsigurl)


class MissingVersion(Exception):
    """The version= line is missing."""


class WatchSyntaxError(Exception):
    """Syntax error in watch file."""


def parse_watch_file(f: Iterable[str]) -> Optional[WatchFile]:
    """Parse a watch file.

    Args:
      f: watch file to parse
    """
    line: Optional[str]
    lines: List[List[str]] = []
    continued: List[str] = []
    for line in f:
        if line.startswith('#'):
            continue
        if not line.strip():
            continue
        if line.rstrip('\n').endswith('\\'):
            continued.append(line.rstrip('\n\\'))
        else:
            continued.append(line)
            lines.append(continued)
            continued = []
    if continued:
        # Hmm, broken line?
        logging.warning('watchfile ended with \\; skipping last line')
        lines.append(continued)
    if not lines:
        return None
    firstline = ''.join(lines.pop(0))
    try:
        key, value = firstline.split('=', 1)
    except ValueError:
        raise MissingVersion()
    if key.strip() != 'version':
        raise MissingVersion()
    version = int(value.strip())
    persistent_options: List[str] = []
    entries: List[Watch] = []
    chunked: List[str]
    for chunked in lines:
        if version > 3:
            chunked = [chunk.lstrip() for chunk in chunked]
        line = ''.join(chunked).strip()
        if not line:
            continue
        opts: Optional[List[str]]
        if line.startswith('opts='):
            if line[5] == '"':
                optend = line.index('"', 6)
                if optend == -1:
                    raise ValueError('Not matching " in %r' % line)
                opts_str = line[6:optend]
                line = line[optend+1:]
            else:
                try:
                    (opts_str, line) = line[5:].split(maxsplit=1)
                except ValueError:
                    opts_str = line[5:]
                    line = None
            opts = opts_str.split(',')
        else:
            opts = None
        if not line:
            if opts:
                persistent_options.extend(opts)
        else:
            try:
                url, line = line.split(maxsplit=1)
            except ValueError:
                url = line
                line = ''
            m = pcre.findall(r'/([^/]*\([^/]*\)[^/]*)$', url)
            if m:
                parts = [m[0]] + line.split(maxsplit=1)
                url = url[:-len(m[0])-1].strip()
            else:
                parts = line.split(maxsplit=2)
            entries.append(Watch(url, *parts, opts=opts))  # type: ignore
    return WatchFile(
        entries=entries, options=persistent_options, version=version)


class WatchEditor(Editor[WatchFile, str]):

    _parsed: WatchFile

    def __init__(
            self, path: str = 'debian/watch', *,
            allow_reformatting: Optional[bool] = None,
            allow_missing: bool = False) -> None:
        super().__init__(
            path, allow_reformatting=allow_reformatting)
        self.allow_missing = allow_missing

    @property
    def watch_file(self) -> WatchFile:
        return self._parsed

    def _nonexistant(self):
        if self.allow_missing:
            return WatchFile([])
        raise

    def _parse(self, content: str) -> WatchFile:
        wf = parse_watch_file(content.splitlines())
        if wf is None:
            return WatchFile([])
        return wf

    def _format(self, parsed: WatchFile) -> str:
        if parsed is None:
            return None
        nf = StringIO()
        parsed.dump(nf)
        return nf.getvalue()


def uscan(wf, package):
    for entry in wf.entries:
        logging.info('entry: %s' % entry)
        for d in entry.discover(package):
            logging.info('  %s' % d)


def main(argv):
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--verbose', action='store_true')
    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG, format='%(message)s')
    else:
        logging.basicConfig(level=logging.INFO, format='%(message)s')
    with open('debian/watch') as f:
        wf = parse_watch_file(f)
    from debian.deb822 import Deb822
    with open('debian/control') as f:
        source = Deb822(f)
    uscan(wf, source['Source'])


if __name__ == '__main__':
    sys.exit(main(sys.argv))



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@debian.org>
# This file is a part of debmutate.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

__all__ = [
    'check_preserve_formatting',
    'check_generated_file',
    'edit_formatted_file',
    'Editor',
    ]


import logging
import os
from typing import Union, Optional, List, TypeVar, Generic


class GeneratedFile(Exception):
    """The specified file is generated."""

    def __init__(self, path: str, template_path=None, template_type=None):
        self.path = path
        self.template_path = template_path
        self.template_type = template_type


class FormattingUnpreservable(Exception):
    """The file is unpreservable."""

    def __init__(self, path, original_contents, rewritten_contents):
        super().__init__(path)
        self.path = path
        self.original_contents = original_contents
        self.rewritten_contents = rewritten_contents

    def diff(self):
        from difflib import unified_diff
        return unified_diff(
            self.original_contents.splitlines(True),
            self.rewritten_contents.splitlines(True),
            fromfile='original', tofile='rewritten')


def check_preserve_formatting(
        rewritten_text: Union[str, bytes], text: Union[str, bytes],
        path: str,
        allow_reformatting: bool = False):
    """Check that formatting can be preserved.

    Args:
      rewritten_text: The rewritten file contents
      text: The original file contents
      path: Path to the file (unused, just passed to the exception)
      allow_reformatting: Whether to allow reformatting
    Raises:
      FormattingUnpreservable: Raised when formatting could not be preserved
    """
    if rewritten_text == text:
        return
    if allow_reformatting:
        return
    raise FormattingUnpreservable(path, text, rewritten_text)


def check_generated_file(path: str) -> None:
    """Check if a file is generated from another file.

    Args:
      path: Path to the file to check
    Raises:
      GeneratedFile: when a generated file is found
    """
    for ext in ['.in', '.m4', '.stub']:
        if os.path.exists(path + ext):
            raise GeneratedFile(path, path + ext)
    DO_NOT_EDIT_SCAN_LINES = 20
    try:
        with open(path, 'rb') as f:
            for i, line in enumerate(f):
                if i > DO_NOT_EDIT_SCAN_LINES:
                    break
                if b"DO NOT EDIT" in line:
                    raise GeneratedFile(path)
                if b"Do not edit!" in line:
                    raise GeneratedFile(path)
                if b'This file is autogenerated' in line:
                    raise GeneratedFile(path)
    except FileNotFoundError:
        return


def edit_formatted_file(
        path: str, original_contents: Union[str, bytes],
        rewritten_contents: Optional[Union[str, bytes]],
        updated_contents: Union[str, bytes],
        allow_generated: bool = False,
        allow_reformatting: bool = False) -> bool:
    """Edit a formatted file.

    Args:
      path: path to the file
      original_contents: The original contents of the file
      rewritten_contents: The contents rewritten with our parser/serializer
      updated_contents: Updated contents rewritten with our parser/serializer
        after changes were made.
      allow_generated: Do not raise GeneratedFile when encountering a generated
        file
      allow_reformatting: Whether to allow reformatting of the file
    """
    if (updated_contents is not None and rewritten_contents is not None
            and type(updated_contents) != type(rewritten_contents)):
        raise TypeError('inconsistent types: {!r}, {!r}'.format(
            type(updated_contents), type(rewritten_contents)))
    if updated_contents in (rewritten_contents, original_contents):
        return False
    if not allow_generated:
        check_generated_file(path)
    try:
        check_preserve_formatting(
            rewritten_contents.strip()  # type: ignore
            if rewritten_contents is not None else None,
            original_contents.strip()
            if original_contents is not None else None, path,
            allow_reformatting=allow_reformatting)
    except FormattingUnpreservable as e:
        if (rewritten_contents is None
                or original_contents is None
                or updated_contents is None):
            raise
        # Run three way merge
        logging.debug(
            'Unable to preserve formatting; falling back to merge3')
        try:
            import merge3
        except ModuleNotFoundError:
            raise e
        if (isinstance(rewritten_contents, bytes)
                and merge3.__version__ < (0, 0, 7)):
            raise e
        m3 = merge3.Merge3(
            rewritten_contents.splitlines(True),
            original_contents.splitlines(True),
            updated_contents.splitlines(True))
        if any([y[0] == 'conflict' for y in m3.merge_regions()]):
            raise
        if isinstance(updated_contents, bytes):
            updated_contents = b''.join(m3.merge_lines())
        else:
            updated_contents = ''.join(m3.merge_lines())
    mode = 'w' + ('b' if isinstance(updated_contents, bytes) else '')
    with open(path, mode) as f:
        f.write(updated_contents)
    return True


T = TypeVar('T')
P = TypeVar('P', str, bytes)


class Editor(Generic[T, P]):
    """Context object for editing a file, preserving formatting."""

    changed_files: List[str]
    _rewritten_content: Optional[P]

    def __init__(
            self, path: str, mode: str = '',
            allow_generated: bool = False,
            allow_reformatting: Optional[bool] = None) -> None:
        self.path = path
        self.mode = mode
        self.allow_generated = allow_generated
        # TODO(jelmer): Don't make this class check the environment
        if allow_reformatting is None:
            allow_reformatting = (
                os.environ.get('REFORMATTING', 'disallow') == 'allow')
        self.allow_reformatting = allow_reformatting

    def _nonexistant(self) -> T:
        raise

    def _parse(self, content: P) -> T:
        """Parse the specified bytestring and returned parsed object."""
        raise NotImplementedError(self._parse)

    def _format(self, parsed: T) -> Optional[P]:
        """Serialize the parsed object."""
        raise NotImplementedError(self._format)

    def __enter__(self):
        try:
            with open(self.path, 'r' + self.mode) as f:
                self._orig_content = f.read()
        except FileNotFoundError:
            self._orig_content = None
            self._parsed = self._nonexistant()
        else:
            self._parsed = self._parse(self._orig_content)
        if self._parsed is not None:
            self._rewritten_content = self._format(self._parsed)
        else:
            self._rewritten_content = None
        return self

    def _updated_content(self) -> Optional[P]:
        if self._parsed is not None:
            return self._format(self._parsed)
        else:
            return None

    def has_changed(self) -> bool:
        """Check if any changes have been made so far."""
        return self._updated_content() not in (
            self._rewritten_content, self._orig_content)

    def __exit__(self, exc_type, exc_val, exc_tb):
        updated_content = self._updated_content()

        if updated_content is None:
            if os.path.exists(self.path):
                os.unlink(self.path)
                self.changed_files = [self.path]
        else:
            self.changed = edit_formatted_file(
                self.path, self._orig_content, self._rewritten_content,
                updated_content, self.allow_generated, self.allow_reformatting)
            if self.changed:
                self.changed_files = [self.path]
            else:
                self.changed_files = []
        return False



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with rules files."""

import os
import re
from typing import Iterator, Optional, List, Union

from .reformatting import Editor


def wildcard_to_re(wildcard: str):
    wc = []
    for c in wildcard:
        if c == '%':
            wc.append('.*')
        else:
            wc.append(re.escape(c))
    return re.compile(''.join(wc))


def matches_wildcard(text: str, wildcard: str) -> bool:
    wc = wildcard_to_re(wildcard)
    return bool(wc.fullmatch(text))


class Rule:
    """A make rule."""

    target: Optional[bytes]
    comands: List[bytes]
    prereq_targets: List[bytes]
    precomment: List[bytes]

    def __init__(self, target: Optional[bytes] = None,
                 commands: Optional[List[bytes]] = None,
                 prereq_targets: Optional[List[bytes]] = None,
                 precomment: Optional[List[bytes]] = None):
        self.precomment = precomment or []
        self.target = target
        self.components = prereq_targets or []
        if self.components:
            self._component_str = b' ' + b' '.join(self.components)
        else:
            self._component_str = b''
        if target:
            self.lines = (
                [b'%s:%s' % (target, self._component_str)]
                + [b'\t' + cmd for cmd in (commands or [])])
        else:
            self.lines = []

    @classmethod
    def _from_first_line(
            cls, firstline: bytes,
            precomment: Optional[List[bytes]] = None) -> 'Rule':
        self = cls(precomment=precomment)
        self.lines = [firstline]
        # TODO(jelmer): What if there are multiple targets?
        self.target, self._component_str = firstline.split(b':', 1)
        self.components = self._component_str.split()
        return self

    def __repr__(self):
        return "<{}({!r})>".format(type(self).__name__, self.target)

    @property
    def targets(self) -> List[bytes]:
        if self.target is None:
            raise AssertionError('no target set')
        return self.target.split(b' ')

    def has_target(self, target: bytes, exact: bool = True) -> bool:
        if exact:
            return target in self.targets
        else:
            return any(
                [matches_wildcard(
                    target.decode(), t.decode()) for t in self.targets])

    def rename_target(self, oldname: bytes, newname: bytes) -> bool:
        # TODO(jelmer): Handle multiple targets
        if self.target == oldname:
            self.target = newname
            firstline = b':'.join([self.target, self._component_str])
            self.lines = [firstline] + self.lines[1:]
            return True
        return False

    def commands(self) -> List[bytes]:
        return [line[1:] for line in self.lines if line.startswith(b'\t')]

    def append_line(self, line: bytes) -> None:
        self.lines.append(line)

    def append_command(self, command: bytes) -> None:
        self.lines.append(b'\t' + command)

    def append_component(self, component: bytes) -> None:
        self.components.append(component)
        self._component_str = b' ' + b' '.join(self.components)
        self.lines[0] = b'%s:%s' % (self.target, self._component_str)

    def remove_component(self, component: bytes) -> None:
        self.components.remove(component)
        self._component_str = b' ' + b' '.join(self.components)
        self.lines[0] = b'%s:%s' % (self.target, self._component_str)

    def dump_lines(self) -> Iterator[bytes]:
        for line in self.precomment:
            yield line + b'\n'
        for line in self.lines:
            yield line + b'\n'

    def __eq__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return False
        return list(self.dump_lines()) == list(other.dump_lines())

    def __bool__(self) -> bool:
        return bool(self.lines)

    def clear(self) -> None:
        self.precomment = []
        self.lines = []

    def _finish(self):
        rest: List[Union[bytes, Rule]] = [self]
        while self.lines and (
                not self.lines[-1] or self.lines[-1].startswith(b'#')):
            rest.insert(1, self.lines.pop(-1))
        return rest


def _is_conditional(line):
    line = line.lstrip(b' ')
    return (
        line.startswith(b'ifeq')
        or line.startswith(b'ifneq')
        or line.startswith(b'else')
        or line.startswith(b'endif')
        or line.startswith(b'include')
        or line.startswith(b'-include'))


def _is_rule(line):
    before, sep, after = line.partition(b':')
    if sep != b':':
        return False
    if b'=' in before:
        return False
    if after.startswith(b'='):
        return False
    return True


class Makefile:
    contents: List[Union[bytes, Rule]]

    def __init__(self, contents: Optional[bytes] = None):
        self.contents = list(contents or [])  # type: ignore

    @classmethod
    def from_path(cls, path: str) -> 'Makefile':
        with open(path, 'rb') as f:
            original_contents = f.read()
        return cls.from_bytes(original_contents)

    def iter_all_rules(self) -> Iterator[Rule]:
        for entry in self.contents:
            if isinstance(entry, Rule):
                yield entry

    def iter_rules(self, target: bytes, exact: bool = True) -> Iterator[Rule]:
        for rule in self.iter_all_rules():
            if rule.has_target(target, exact):
                yield rule

    def get_variable(self, desired_key: bytes) -> bytes:
        for line in self.contents:
            if not isinstance(line, bytes):
                continue
            m = re.fullmatch(
                br'(export\s)?([A-Za-z0-9_]+)\s*[:?]?=\s*(.*)', line)
            if not m:
                continue
            if m.group(2).strip() == desired_key:
                return m.group(3).strip()
        raise KeyError(desired_key)

    @classmethod
    def from_bytes(cls, contents):
        mf = cls()
        keep = b''
        rule = None
        joinedlines = []
        for line in contents.splitlines():
            line = keep + line
            keep = b''
            if line.endswith(b'\\'):
                keep = line + b'\n'
                continue
            joinedlines.append(line)

        if keep:
            # file ends with continuation line..
            joinedlines.append(keep)

        for line in joinedlines:
            if line.startswith(b'\t') and rule:
                rule.append_line(line)
            elif _is_conditional(line) or line.lstrip(b' ').startswith(b'#'):
                if rule:
                    rule.append_line(line)
                else:
                    mf.contents.append(line)
            elif _is_rule(line):
                if rule:
                    mf.contents.extend(rule._finish())
                precomment: List[bytes] = []
                while (len(mf.contents) > 1
                        and isinstance(mf.contents[-1], bytes)
                        and mf.contents[-1].startswith(b'#')):
                    precomment.insert(0, mf.contents.pop(-1))  # type: ignore
                rule = Rule._from_first_line(line, precomment=precomment)
            elif not line.strip():
                if rule:
                    rule.append_line(line)
                else:
                    mf.contents.append(line)
            else:
                if rule:
                    mf.contents.extend(rule._finish())
                rule = None
                mf.contents.append(line)

        if rule:
            mf.contents.extend(rule._finish())

        return mf

    def dump_lines(self):
        lines: List[bytes] = []
        contents = self.contents[:]
        while contents and contents[-1] == b'':
            del contents[-1]
        for entry in contents:
            if isinstance(entry, Rule):
                if entry.lines == []:
                    if lines[-1] == b'\n':
                        lines.pop(-1)
                else:
                    lines.extend(entry.dump_lines())
            else:
                lines.append(entry + b'\n')
        return lines

    def dump(self):
        return b''.join(self.dump_lines())

    def add_rule(self, target, components=None, precomment=None) -> Rule:
        if self.contents:
            if isinstance(self.contents[-1], Rule):
                self.contents.append(b'')
        if isinstance(target, list):
            target = b' '.join(target)
        line = b'%s:' % target
        if components:
            line += b' ' + b' '.join(components)
        rule = Rule._from_first_line(line, precomment=precomment)
        self.contents.append(rule)
        return rule

    def mark_phony(self, rule):
        for r in self.iter_rules(b'.PHONY'):
            if rule in r.components:
                return
        try:
            phony_rule = list(self.iter_rules(b'.PHONY'))[-1]
        except IndexError:
            self.add_rule(b'.PHONY', [rule])
        else:
            phony_rule.append_component(rule)

    def add_phony(self, rule):
        try:
            phony_rule = list(self.iter_rules(b'.PHONY'))[-1]
        except IndexError:
            return

        phony_rule.append_component(rule)

    def drop_phony(self, rule):
        for r in self.iter_rules(b'.PHONY'):
            if rule in r.components:
                r.remove_component(rule)
            if not r.components:
                r.clear()


class MakefileEditor(Editor[Makefile, bytes]):

    def __init__(self, path):
        super().__init__(path, mode='b')

    def _parse(self, content):
        return Makefile.from_bytes(content)

    def _format(self, parsed):
        return parsed.dump()

    @property
    def makefile(self):
        return self._parsed


class RulesEditor(MakefileEditor):

    def __init__(self, path='debian/rules'):
        super().__init__(path)

    def legacy_update(self, command_line_cb=None, global_line_cb=None,
                      rule_cb=None, makefile_cb=None,
                      drop_related_comments=False):
        """Update a debian/rules file.

        Args:
          command_line_cb: Callback to call on every rule command line
          global_line_cb: Callback to call on every global line
          rule_cb: Callback to call on every rule
        Returns:
          boolean indicating whether any changes were made
        """
        newcontents: List[Union[bytes, Rule]] = []
        for entry in self.makefile.contents:
            if isinstance(entry, Rule):
                rule = entry
                newlines = []
                for line in list(rule.lines[1:]):
                    if line.startswith(b'\t'):
                        ret = line[1:]
                        if callable(command_line_cb):
                            ret = command_line_cb(ret, rule.target)
                        elif isinstance(command_line_cb, list):
                            for fn in command_line_cb:
                                ret = fn(ret, rule.target)
                        if isinstance(ret, bytes):
                            newlines.append(b'\t' + ret)
                        elif isinstance(ret, list):
                            for otherl in ret:
                                newlines.append(b'\t' + otherl)
                        else:
                            raise TypeError(ret)
                    else:
                        newlines.append(line)

                rule.lines = [rule.lines[0]] + newlines

                if rule_cb:
                    rule_cb(rule)
                if rule:
                    newcontents.append(rule)
                else:
                    if newcontents and newcontents[-1] == b'':
                        newcontents.pop(-1)
            else:
                line = entry
                if global_line_cb:
                    line = global_line_cb(line)
                if line is None:
                    if drop_related_comments:
                        while newcontents and newcontents[-1].startswith(b'#'):
                            del newcontents[-1]
                    if newcontents and not newcontents[-1]:
                        del newcontents[-1]
                    # TODO(jelmer): If there is no preceding whitespace, drop
                    # next line if it's empty?
                elif isinstance(line, list):
                    newcontents.extend(line)
                elif isinstance(line, bytes):
                    newcontents.append(line)
                else:
                    raise TypeError(line)

        self.makefile.contents = newcontents
        if makefile_cb:
            makefile_cb(self.makefile)
        if self.has_changed():
            discard_pointless_overrides(self.makefile)
            return True
        else:
            return False


def discard_pointless_overrides(makefile, ignore_comments=False):
    for rule in makefile.iter_all_rules():
        discard_pointless_override(
            makefile, rule, ignore_comments=ignore_comments)


def discard_pointless_override(makefile, rule, ignore_comments=False):
    if not rule.target.startswith(b'override_'):
        return
    command = rule.target[len(b'override_'):]
    if ignore_comments:
        effective_lines = [
            line for line in rule.lines[1:]
            if line.split(b'#', 1)[0].strip()]
    else:
        effective_lines = [
            line for line in rule.lines[1:]
            if line.strip()]
    if effective_lines != [b'\t' + command]:
        return
    if rule.components:
        return
    rule.clear()
    makefile.drop_phony(rule.target)


def update_rules(command_line_cb=None, global_line_cb=None,
                 rule_cb=None,
                 makefile_cb=None, path='debian/rules',
                 drop_related_comments=False):
    """Update a debian/rules file.

    Args:
      command_line_cb: Callback to call on every rule command line
      global_line_cb: Callback to call on every global line
      rule_cb: Callback to call on every rule
      path: Path to the debian/rules file to edit
    Returns:
      boolean indicating whether any changes were made
    """
    if not os.path.exists(path):
        return False
    with RulesEditor(path) as updater:
        updater.legacy_update(
            command_line_cb=command_line_cb,
            global_line_cb=global_line_cb,
            rule_cb=rule_cb,
            makefile_cb=makefile_cb,
            drop_related_comments=drop_related_comments)
    return updater.changed


def dh_invoke_add_with(line, with_argument):
    """Add a particular value to a with argument."""
    if with_argument in line:
        return line
    if b' --with' not in line:
        return line + b' --with=' + with_argument
    return re.sub(
        b'([ \t])--with([ =])([^ \t]+)', b'\\1--with\\2\\3,' + with_argument,
        line)


def dh_invoke_get_with(line: bytes) -> List[str]:
    ret: List[str] = []
    for m in re.finditer(b'[ \t]--with[ =]([^ \t]+)', line):
        ret.extend(m.group(1).decode('utf-8').split(','))
    return ret


def dh_invoke_drop_with(line, with_argument):
    """Drop a particular value from a with argument."""
    if with_argument not in line:
        return line
    # It's the only with argument
    line = re.sub(
        b"[ \t]--with[ =]" + with_argument + b"( .+|)$",
        b"\\1", line)
    # It's at the beginning of the line
    line = re.sub(
        b"([ \t])--with([ =])" + with_argument + b",",
        b"\\1--with\\2", line)
    # It's somewhere in the middle or the end
    line = re.sub(
        b"([ \t])--with([ =])(.+)," + with_argument + b"([ ,])",
        b"\\1--with\\2\\3\\4", line)
    # It's at the end
    line = re.sub(
        b"([ \t])--with([ =])(.+)," + with_argument + b"$",
        b"\\1--with\\2\\3", line)
    return line


def dh_invoke_drop_argument(line, argument):
    """Drop a particular argument from a dh invocation."""
    if argument not in line:
        return line
    line = re.sub(b'[ \t]+' + argument + b'$', b'', line)
    line = re.sub(b'([ \t])' + argument + b'[ \t]', b'\\1', line)
    return line


def dh_invoke_replace_argument(line, old, new):
    if old not in line:
        return line
    line = re.sub(b'([ \t])' + old + b'$', b'\\1' + new, line)
    line = re.sub(
        b'([ \t])' + old + b'([ \t])', b'\\1' + new + b'\\2', line)
    return line


def check_cdbs(path='debian/rules'):
    if not os.path.exists(path):
        return False
    with open(path, 'rb') as f:
        for line in f:
            if line.lstrip(b'-').startswith(b'include /usr/share/cdbs/'):
                return True
    return False



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij
# This file is a part of debmutate.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with deb822 files."""

__all__ = [
    'dump_paragraphs',
    'reformat_deb822',
    'ChangeConflict',
    'Deb822Editor',
    'parse_deb822_file',
    'Deb822Paragraph',
]

from io import BytesIO
from typing import List, Optional, Union

from debian.deb822 import Deb822
from debian._deb822_repro.parsing import (
    parse_deb822_file,
    Deb822ParagraphElement as Deb822Paragraph,
    Deb822FileElement as Deb822File,
)

from .reformatting import (
    Editor,
)


# Urgh; this is bad form as a library, but the alternative is
# silently discarding comments.
Deb822Paragraph._discard_comments_on_read = False  # type: ignore


def dump_paragraphs(paragraphs: Union[Deb822File, List[Deb822]]) -> bytes:
    """Dump a set of deb822 paragraphs to a file.

    Args:
      paragraphs: iterable over paragraphs
    Returns:
      formatted text (as bytes)
    """
    outf = BytesIO()
    if hasattr(paragraphs, 'dump'):
        # deb822_repro
        paragraphs.dump(outf)
    else:
        first = True
        for paragraph in paragraphs:
            if paragraph:
                if not first:
                    outf.write(b'\n')
                paragraph.dump(fd=outf)
                first = False
    return outf.getvalue()


def reformat_deb822(contents: bytes) -> bytes:
    """Check whether it's possible to preserve a control file.

    Args:
      contents: Original contents
    Returns:
      New contents
    """
    return dump_paragraphs(parse_deb822_file(BytesIO(contents)))


class ChangeConflict(Exception):
    """Indicates that a proposed change didn't match what was found.
    """

    def __init__(self, para_key, field, expected_old_value, actual_old_value,
                 new_value):
        self.paragraph_key = para_key
        self.field = field
        self.expected_old_value = expected_old_value
        self.actual_old_value = actual_old_value
        self.new_value = new_value

    def __repr__(self):
        return ("%s(para_key=%r, field=%r, expected_old_value=%r, "
                "actual_old_value=%r, new_value=%r)" % (
                    type(self).__name__,
                    self.paragraph_key, self.field, self.expected_old_value,
                    self.actual_old_value, self.new_value))


class Deb822Editor(Editor[List[Deb822Paragraph], bytes]):
    """Update the contents of a Deb822-style file.

    """

    def __init__(self, path: str, allow_generated: bool = False,
                 allow_reformatting: Optional[bool] = None,
                 allow_missing: bool = False,
                 accept_files_with_error_tokens: bool = False) -> None:
        super().__init__(
            path, allow_generated=allow_generated,
            allow_reformatting=allow_reformatting,
            mode='b')
        self.allow_missing = allow_missing
        self.accept_files_with_error_tokens = accept_files_with_error_tokens

    def apply_changes(self, changes, resolve_conflict=None):
        """Apply a set of changes to this deb822 instance.

        Args:
          changes: dict mapping paragraph types to
              list of (field_name, old_value, new_value)
            resolve_conflict: callback that receives
                (para_key, field, actual_old_value, template_old_value,
                 actual_new_value) and returns a new template value
        """
        def _default_resolve_conflict(
                para_key, field, actual_old_value,
                template_old_value, actual_new_value):
            raise ChangeConflict(
                para_key, field, actual_old_value, template_old_value,
                actual_new_value)

        if resolve_conflict is None:
            resolve_conflict = _default_resolve_conflict

        changes = dict(changes.items())
        for paragraph in self.paragraphs:
            for item_ in list(paragraph.items()):
                item = (str(item_[0]), str(item_[1]))
                for key, old_value, new_value in changes.pop(item, []):
                    if paragraph.get(key) != old_value:
                        new_value = resolve_conflict(
                            item, key, old_value, paragraph.get(key),
                            new_value)
                    if new_value is None:
                        del paragraph[key]
                    else:
                        paragraph[key] = new_value
        # Add any new paragraphs that weren't processed earlier
        for key, p in changes.items():
            paragraph = Deb822Paragraph.new_empty_paragraph()
            for (field, old_value, new_value) in p:
                if old_value is not None:
                    new_value = resolve_conflict(
                        key, field, old_value, None, new_value)
                if new_value is None:
                    continue
                paragraph[field] = new_value
            self.paragraphs.append(paragraph)

    def _parse(self, content):
        return parse_deb822_file(
            content.splitlines(True),
            accept_files_with_error_tokens=self.accept_files_with_error_tokens)

    @property
    def paragraphs(self) -> List[Deb822Paragraph]:
        return self._parsed

    def _format(self, paragraphs):
        return dump_paragraphs(paragraphs)

    def _nonexistant(self):
        if self.allow_missing:
            return parse_deb822_file([])
        raise

    def sort_paragraphs(self, sort_key, skip=0):
        sortable = list(self.paragraphs)[skip:]
        for i, p in reversed(list(enumerate(self.paragraphs))):
            if i < skip:
                continue
            self.paragraphs.remove(p)
        for p in sorted(sortable, key=sort_key):
            self.paragraphs.append(p)
